from copy import deepcopy

import cv2
from scipy.signal import argrelmax

from mower.solvers.navigation import NavigationSolver
from mower.utils import config
from mower.utils.character_recognize import match_avatar
from mower.utils.graph import SceneGraphSolver
from mower.utils.image import cropimg, loadres, thres2
from mower.utils.log import logger
from mower.utils.scene import Scene
from mower.utils.tile_pos import Calc, find_level
from mower.utils.vector import sa

from .mixin import FightMixin


class AutoFight(SceneGraphSolver, FightMixin):
    def run(
        self,
        level_name: str,
        opers: list,
        actions: list,
        groups: list = [],
    ):
        """
        Args:
            level_name: 关卡名
            opers: 干员列表
            actions: 行动列表
            groups: 干员组列表
        """
        logger.info("Start: 自动战斗")
        logger.info("地图坐标计算：https://github.com/yuanyan3060/Arknights-Tile-Pos")
        self.level_name = level_name
        level = find_level(level_name, None)
        if level is None:
            level = find_level(None, level_name)
        if level is None:
            logger.error(f"找不到关卡：{level_name}")
            return False
        self.calc = Calc(1920, 1080, level)
        self.actions = deepcopy(actions)  # 行动列表
        self.opers = opers  # 干员列表
        self.groups = deepcopy(groups)  # 干员组
        self.speed = 1  # 一倍/二倍速
        self.loading = True  # 等待加载
        self.playing = True  # 暂停/继续
        self.Bullet_Time = False  # 是否选中干员进入1/5速度
        self.condition_satified = False  # 前置条件是否满足
        self.operators = {}  # 可部署的干员
        self.location = {}  # 干员部署坐标
        self.post_delay = 0  # 行动后等待时间
        self.watching = {}  # 需要开技能的干员
        self.replacements = {}  # 将干员组的名字替换为对应干员
        self.success = True
        super().run()
        return self.success

    def change_group(self):
        for group in self.groups:
            for op in group["opers"]:
                self.replacements[op["name"]] = group["name"]
                op["name"] = group["name"]
                self.opers.append(op)

    def update_watching(self):
        for op in self.opers:
            name = op["name"]
            self.watching[name] = deepcopy(op)
            self.watching[name]["location"] = None

    def kills(self) -> int:
        "获取击杀数"
        img = cropimg(config.recog.gray, ((800, 30), (950, 60)))
        img = thres2(img, 127)
        sep = loadres("fight/kills_separator", True)
        result = cv2.matchTemplate(img, sep, cv2.TM_SQDIFF_NORMED)
        min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(result)
        x = min_loc[0] + 799
        kills = self.number(((800, 30), (x, 60)), 28, 127)
        logger.debug(kills)
        return kills

    def check_condition(self) -> bool:
        if "kills" in self.action and self.action["kills"] > self.kills():
            return False
        if "costs" in self.action and self.action["costs"] > self.cost():
            return False
        if "cooling" in self.action and self.action["cooling"] > self.CD():
            return False
        if "cost_changes" in self.action:
            s = self.cost()
            while self.cost() - s < self.action["cost_changes"]:
                self.sleep(0.5)
        if "pre_delay" in self.action:
            self.sleep(self.action["pre_delay"] / 1000)
        if "post_delay" in self.action:
            self.pre_delay = self.action["post_delay"] / 1000
        if self.Bullet_Time:
            self.tap((400, 100))
            self.sleep(0.3)
            self.Bullet_Time = False
        return True

    def update_operators(self):
        "识别可部署的干员"
        if config.conf.avatar_recog_pause:
            self.pause()
        config.recog.update()
        y = 887
        img = cropimg(config.recog.gray, ((0, y), (1920, 905)))
        threshold = 0.7
        c = loadres("fight/c", True)
        mask = loadres("fight/c_mask", True)
        result = cv2.matchTemplate(img, c, cv2.TM_CCOEFF_NORMED, None, mask)[0]
        op = []
        for i in argrelmax(result, order=50)[0]:
            if result[i] > threshold:
                op.append(i)
        segment = []
        cost = []
        self.operators = {}
        for x in op:
            # 看最下方条的颜色判断是否正在转CD
            bar_scope = sa(((-20, 187), (10, 190)), (x, y))
            img = cropimg(config.recog.img, bar_scope)
            img = cv2.inRange(img, (1, 0, 0), (3, 255, 255))
            count = cv2.countNonZero(img)
            logger.debug(count)
            if count > 50:
                continue

            scope = sa(((-58, 53), (42, 153)), (x, y))
            segment.append(scope)
            cost_scope = sa(((-13, 19), (30, 44)), (x, y))
            cost.append(self.number(cost_scope, 25, 80))
        operators = match_avatar(config.recog.img, segment)
        for idx, (name, scope) in enumerate(operators):
            name = self.replacements[name] if name in self.replacements else name
            self.operators[name] = {"scope": scope, "cost": cost[idx]}
        logger.debug(self.operators)
        if config.conf.avatar_recog_pause:
            self.play()

    def complete_action(self):
        "完成当前行动"
        self.actions.pop(0)
        self.condition_satified = False
        if self.post_delay > 0:
            self.sleep_post_delay()

    def play(self):
        logger.info("继续")
        while not self.playing:
            self.toggle_play()

    def pause(self):
        logger.info("暂停")
        while self.playing:
            self.toggle_play()

    def action_skill(self):
        "使用action中指定位置的技能"
        if flag := ("name" in self.action):
            name = self.action["name"]
            x, y = self.location[name]

        else:
            x, y = self.action["location"]
        if self.skill_ready(x, y) or self.skill_stop(x, y):
            self.use_skill(x, y)
            if flag and self.watching[self.action["name"]]["skill_usage"] == 2:
                self.watching[self.action["name"]]["skill_times"] -= 1
                if self.watching[self.action["name"]]["skill_times"] <= 0:
                    del self.watching[self.action["name"]]

            self.complete_action()

    def transition(self):
        if (scene := self.scene()) == Scene.OPERATOR_FIGHT:
            if self.loading:
                self.pause()
                self.loading = False
                self.change_group()
                self.update_operators()
                self.update_watching()
                self.play()
                return

            for w, d in self.watching.items():
                if w not in self.location or d["skill_usage"] == 0:
                    continue
                x, y = self.location[w]
                if self.skill_ready(x, y):
                    if self.use_skill(x, y):
                        if d["skill_usage"] == 2:
                            d["skill_times"] -= 1
                            if d["skill_times"] <= 0:
                                del self.watching[w]
                        return
            if self.action is None:
                return
            if not self.condition_satified:
                self.condition_satified = self.check_condition()
                return
            logger.debug(self.action)
            if self.action["type"] in ["SpeedUp", "二倍速"]:
                self.toggle_speed()
            elif self.action["type"] in ["Deploy", "部署"]:
                self.deploy()
            elif self.action["type"] in ["Retreat", "撤退"]:
                self.withdraw()
            elif self.action["type"] in ["Skill", "技能"]:
                self.action_skill()
            elif self.action["type"] in ["SkillDaemon", "摆完挂机"]:
                self.complete_action()
            elif self.action["type"] in ["SkillUsage", "技能用法"]:
                self.watching[self.action["name"]]["skill_usage"] = self.action[
                    "skill_usage"
                ]
                self.complete_action()
            elif self.action["type"] in ["BulletTime", "子弹时间"]:
                self.bullet_time()

            elif self.action["type"] in ["Output", "打印"]:
                self.complete_action()

        elif scene == Scene.OPERATOR_FAILED:
            logger.warning("行动失败，请检查干员/练度")
            self.success = False
            return True

        elif scene == Scene.OPERATOR_FINISH:
            logger.info("行动结束")
            return True

        elif scene == Scene.CONFIRM:
            logger.warning("被顶号")
            self.success = False
            return True

        elif scene in self.waiting_scene:
            self.waiting_solver()
        elif scene == Scene.OPERATOR_SELECT:
            self.tap((1655, 781))
        elif scene == Scene.STORY:
            self.tap_element("story_skip")
        elif scene == Scene.STORY_SKIP:
            self.tap_element("story_skip_confirm_dialog", x_rate=0.94)
        else:
            NavigationSolver().run(self.level_name, mode="copy")
