import cv2
import numpy as np
from scipy.signal import argrelmax
from skimage.metrics import structural_similarity

from mower.utils import config
from mower.utils import typealias as tp
from mower.utils.image import cropimg, loadres, thres2
from mower.utils.log import logger
from mower.utils.scene import Scene
from mower.utils.vector import sa, va


class FightMixin:
    @property
    def action(self):
        "下一步行动"
        if len(self.actions) > 0:
            return self.actions[0]
        return None

    def sleep_post_delay(self):
        self.sleep(self.post_delay)
        self.post_delay = 0

    def number(self, scope: tp.Scope, height: int, thres: int) -> int:
        "数字识别"
        return config.recog.num.number_int("secret_front", scope, height, thres)

    def cost(self) -> int:
        "获取部署费用"
        cost = self.number(((1800, 745), (1920, 805)), 52, 200)
        logger.debug(cost)
        return cost

    def play(self):
        logger.info("继续")
        while not self.playing and self.scene() == Scene.OPERATOR_FIGHT:
            self.toggle_play()

    def pause(self):
        logger.info("暂停")
        while self.playing and self.scene() == Scene.OPERATOR_FIGHT:
            self.toggle_play()

    def toggle_play(self):
        "切换暂停与继续"
        self.tap((1800, 80), interval=0.5)
        img = cropimg(config.recog.gray, ((740, 480), (1180, 665)))
        img = thres2(img, 250)
        res = loadres("fight/pause", True)
        ssim = structural_similarity(img, res)
        logger.debug(ssim)
        self.playing = ssim <= 0.9

    def toggle_speed(self):
        target = 1 if self.speed == 2 else 2
        logger.info(f"切换至{target}倍速")
        config.device.tap((1650, 80))
        self.speed = target
        self.complete_action()

    def select(self, x: int, y: int):
        "选中干员"
        pos = self.calc.get_character_screen_pos(x, y, False, False)
        pos = int(pos.x), int(pos.y)
        config.device.tap(pos)

    def withdraw(self):
        "撤下干员"
        if "name" in self.action:
            name = self.action["name"]
            x, y = self.location[name]
            self.location[name] = None
        else:
            x, y = self.action["location"]
        self.select(x, y)
        self.sleep(0.5)
        pos = self.calc.get_with_draw_screen_pos(x, y)
        pos = int(pos.x), int(pos.y)
        config.device.tap(pos)
        self.sleep(0.5)
        self.operators = {}
        self.complete_action()

    def use_skill(self, x, y):
        "开技能"
        self.select(x, y)
        self.sleep(0.4)
        pos = self.calc.get_skill_screen_pos(x, y)
        pos = int(pos.x), int(pos.y)
        self.tap(pos)
        self.sleep(0.4)
        if self.find("fight/attack"):
            self.tap((20, 20))
            return False
        return True

    def bullet_time(self):
        "选中干员进入1/5速度"
        if "name" in self.action:
            name = self.action["name"]
            x, y = self.location[name]
            self.location[name] = None
        else:
            x, y = self.action["location"]
        self.select(x, y)
        self.Bullet_Time = True
        self.complete_action()

    def drag_success(self):
        try:
            img = thres2(config.recog.gray, 200)
            contours, _ = cv2.findContours(
                img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
            )
            rect = [cv2.boundingRect(c) for c in contours]
            left = min(x for x, y, w, h in rect if w > 20 and h > 20)
            right = max(x + w for x, y, w, h in rect if w > 20 and h > 20)
            top = min(y for x, y, w, h in rect if w > 20 and h > 20)
            bottom = max(y + h for x, y, w, h in rect if w > 20 and h > 20)
            width = right - left
            height = bottom - top
            logger.debug(f"{width=} {height=}")
            return width < 900 and height < 850
        except Exception:
            return False

    def CD(self) -> int:
        "获取进入CD的干员数量"
        config.recog.update()
        y = 887
        img = cropimg(config.recog.gray, ((0, y), (1920, 905)))
        threshold = 0.7
        c = loadres("fight/c", True)
        mask = loadres("fight/c_mask", True)
        result = cv2.matchTemplate(img, c, cv2.TM_CCOEFF_NORMED, None, mask)[0]
        op = []
        for i in argrelmax(result, order=50)[0]:
            if result[i] > threshold:
                op.append(i)
        self.operators = {}
        number = 0
        for x in op:
            # 看最下方条的颜色判断是否正在转CD
            bar_scope = sa(((-20, 187), (10, 190)), (x, y))
            img = cropimg(config.recog.img, bar_scope)
            img = cv2.inRange(img, (1, 0, 0), (3, 255, 255))
            count = cv2.countNonZero(img)
            logger.debug(count)
            if count > 50:
                number += 1
        return number

    def skill_ready(self, x: int, y: int) -> bool:
        """指定坐标的干员技能是否可以开启

        Args:
            x: 横坐标
            y: 纵坐标
        """
        try:
            skill_ready = loadres("fight/skill_ready", True)
            pos = self.calc.get_character_screen_pos(x, y, False, False)
            pos = int(pos.x), int(pos.y)
            img = cropimg(config.recog.gray, sa(((-15, -168), (20, -128)), pos))
            mask = np.ones_like(skill_ready, dtype=np.float32)
            mask[8:17, :] = 0.3
            result = cv2.matchTemplate(
                img, skill_ready, cv2.TM_CCOEFF_NORMED, mask=mask
            )
            min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(result)
            logger.debug(f"{max_val=} {max_loc=}")
            return max_val >= 0.65
        except Exception:
            return False

    def skill_stop(self, x: int, y: int) -> bool:
        """指定坐标的干员技能是否可以关闭

        Args:
            x: 横坐标
            y: 纵坐标
        """
        try:
            skill_stop = loadres("fight/skill_stop")
            h, w, _ = skill_stop.shape
            pos = self.calc.get_character_screen_pos(x, y, False, False)
            pos = int(pos.x), int(pos.y)
            img = cropimg(config.recog.img, sa(((-15, -168), (15, -138)), pos))
            result = cv2.matchTemplate(img, skill_stop, cv2.TM_SQDIFF_NORMED)
            min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(result)
            logger.debug(f"{min_val=} {min_loc=}")
            return min_val <= 0.2
        except Exception:
            return False

    def deploy(self):
        "部署干员"
        name = self.action["name"]
        if name not in self.operators:
            self.update_operators()
            return
        if self.cost() < self.operators[name]["cost"]:
            return
        start = self.get_pos(self.operators[name]["scope"])
        top = va(start, (0, -100))
        x, y = self.action["location"]
        pos = self.calc.get_character_screen_pos(x, y, True, False)
        pos = round(pos.x), round(pos.y)
        direction = self.action["direction"]
        logger.info(f"在({x}, {y})部署{name}，方向为{direction}")
        if direction in ["Left", "左"]:
            dir = (-200, 0)
        elif direction in ["Right", "右"]:
            dir = (200, 0)
        elif direction in ["Up", "上"]:
            dir = (0, -200)
        else:
            dir = (0, 200)
        dir = va(pos, dir)
        config.device.swipe_ext([start, top, pos], [200, 1000], up_wait=400)
        if direction in ["None", "无"]:
            self.operators = {}
            self.complete_action()
            self.location[name] = x, y
            return

        for _ in range(5):
            if not (flag := self.drag_success()):
                self.sleep(0.1)
            else:
                break

        if flag:
            config.device.swipe_ext([pos, dir], [200])
            self.sleep(0.5)
            if self.drag_success():
                self.ctap(va(pos, (-300, -300)), 0.5, config.screenshot_avg / 1000)
                return
            self.operators = {}
            self.complete_action()
            self.location[name] = x, y
        else:
            self.ctap(va(pos, (-300, -300)), 0.5, config.screenshot_avg / 1000)
